'''
    Contains mid layer functions that will be used by the IMC binding and are not directly
    related to the parsing/extract process from the XML file. (lower layers).

    All functions should be "universal" across different versions of IMC.

    Should be a standalone module that depends only on python modules and be
    used as a dependency of higher layers such as _base_templates.py and network.py.

    Obs: There are some "assumptions" here. For example, we assume that the message class
    will have an attribute called '_Attributes', which will have a field called 'fields'.
    Although '_Attributes' is programmatically generated, the field 'fields' is not. It
    is read from IMC definition.
'''

import ifaddr as _ifaddr
import ipaddress as _ipaddress

import struct as _struct
import asyncio as _asyncio

# This has to be hardcoded =/
# be = Big Endian, le = Little Endian

pack_functions_big = {
'int8_t': _struct.Struct('>b').pack,
'uint8_t': _struct.Struct('>B').pack,
'int16_t': _struct.Struct('>h').pack,
'uint16_t': _struct.Struct('>H').pack,
'int32_t': _struct.Struct('>i').pack,
'uint32_t': _struct.Struct('>I').pack,
'int64_t': _struct.Struct('>q').pack,
'fp32_t': _struct.Struct('>f').pack,
'fp64_t': _struct.Struct('>d').pack,
'rawdata': lambda x : _struct.Struct('>H').pack(len(x)) + x,
'plaintext': lambda x : _struct.Struct('>H').pack(len(x)) + x.encode(encoding = 'ascii', errors='surrogateescape'),
'message': lambda x : x.pack(is_field_message=True, is_big_endian=True),
'message-list': lambda x : b''.join([_struct.Struct('>H').pack(len(x)), *[m.pack(is_field_message=True, is_big_endian=True) for m in x]]),
'header': _struct.Struct('>HHHdHBHB').pack # special "type"
}

pack_functions_little = {
'int8_t': _struct.Struct('<b').pack,
'uint8_t': _struct.Struct('<B').pack,
'int16_t': _struct.Struct('<h').pack,
'uint16_t': _struct.Struct('<H').pack,
'int32_t': _struct.Struct('<i').pack,
'uint32_t': _struct.Struct('<I').pack,
'int64_t': _struct.Struct('<q').pack,
'fp32_t': _struct.Struct('<f').pack,
'fp64_t': _struct.Struct('<d').pack,
'rawdata': lambda x : _struct.Struct('<H').pack(len(x)) + x,
'plaintext': lambda x : _struct.Struct('<H').pack(len(x)) + x.encode(encoding = 'ascii', errors='surrogateescape'),
'message': lambda x : x.pack(is_field_message=True, is_big_endian=False),
'message-list': lambda x : b''.join([_struct.Struct('<H').pack(len(x)), *[m.pack(is_field_message=True, is_big_endian=False) for m in x]]),
'header': _struct.Struct('<HHHdHBHB').pack # special "type"
}

unpack_functions_big = {
'int8_t': lambda x : (_struct.Struct('>b').unpack(x[:1])[0], 1),
'uint8_t': lambda x : (_struct.Struct('>B').unpack(x[:1])[0], 1),
'int16_t': lambda x : (_struct.Struct('>h').unpack(x[:2])[0], 2),
'uint16_t': lambda x : (_struct.Struct('>H').unpack(x[:2])[0], 2),
'int32_t': lambda x : (_struct.Struct('>i').unpack(x[:4])[0], 4),
'uint32_t': lambda x : (_struct.Struct('>I').unpack(x[:4])[0], 4),
'int64_t': lambda x : (_struct.Struct('>q').unpack(x[:8])[0], 8),
'fp32_t': lambda x : (_struct.Struct('>f').unpack(x[:4])[0], 4),
'fp64_t': lambda x : (_struct.Struct('>d').unpack(x[:8])[0], 8),
'rawdata': lambda x : (x[2:2 + _struct.Struct('>H').unpack(x[:2])[0]], 2 + _struct.Struct('>H').unpack(x[:2])[0]),
'plaintext': lambda x : (x[2:2 + int.from_bytes(x[:2], byteorder='big')].decode(encoding = 'ascii', errors='surrogateescape'), 2 + int.from_bytes(x[:2], byteorder='big')),
'message': None,
'message-list': None,
'header': lambda x : (_struct.Struct('>HHHdHBHB').unpack(x[:20]), 20), # special "type"
}

unpack_functions_little = {
'int8_t': lambda x : (_struct.Struct('<b').unpack(x[:1])[0], 1),
'uint8_t': lambda x : (_struct.Struct('<B').unpack(x[:1])[0], 1),
'int16_t': lambda x : (_struct.Struct('<h').unpack(x[:2])[0], 2),
'uint16_t': lambda x : (_struct.Struct('<H').unpack(x[:2])[0], 2),
'int32_t': lambda x : (_struct.Struct('<i').unpack(x[:4])[0], 4),
'uint32_t': lambda x : (_struct.Struct('<I').unpack(x[:4])[0], 4),
'int64_t': lambda x : (_struct.Struct('<q').unpack(x[:8])[0], 8),
'fp32_t': lambda x : (_struct.Struct('<f').unpack(x[:4])[0], 4),
'fp64_t': lambda x : (_struct.Struct('<d').unpack(x[:8])[0], 8),
'rawdata': lambda x : (x[2:2 + _struct.Struct('<H').unpack(x[:2])[0]], 2 + _struct.Struct('<H').unpack(x[:2])[0]),
'plaintext': lambda x : (x[2:2 + int.from_bytes(x[:2], byteorder='little')].decode(encoding = 'ascii', errors='surrogateescape'), 2 + int.from_bytes(x[:2], byteorder='little')),
'message': None,
'message-list': None,
'header': lambda x : (_struct.Struct('<HHHdHBHB').unpack(x[:20]), 20) # special "type"
}

crc16_ibm_table_uint = [
      0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241,
      0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
      0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40,
      0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
      0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40,
      0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
      0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641,
      0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
      0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240,
      0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
      0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41,
      0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
      0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41,
      0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
      0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640,
      0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
      0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240,
      0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
      0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41,
      0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
      0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41,
      0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
      0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640,
      0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
      0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241,
      0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
      0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40,
      0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
      0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40,
      0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
      0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641,
      0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040]

def CRC16IMB(message : bytes) -> int:
    '''Return value should be an uint, even though it is annoted with int (there is no uint in Python)'''
    result = 0
    for m in message:
            result = (result >> 8) ^ crc16_ibm_table_uint[((result ^ m) & 0xFF)]
    return result

def get_initial_IP() -> int:
    '''Returns the 1st non-localhost IPv4 if it exists. Else, returns localhost
    
    obs: Returns as int.
    '''
    ipv4_list = [IP.ip for interface in _ifaddr.get_adapters() for IP in interface.ips if isinstance(IP.ip, str)]
    
    for ip in ipv4_list:
        if ip != '127.0.0.1':
            return int(_ipaddress.IPv4Address(ip))
    return int(_ipaddress.IPv4Address('127.0.0.1'))

async def _async_wrapper(func, *args):
    return func(*args)

class base_IO_interface:
    '''
        An 'abstract'* class that describes the basic implementation of an I/O interface.

        * Not really abstract as in Java, but consider it so. 
        I will not use abc and its decorators.
    '''
    __slots__ = ['_input', '_output', '_o', '_i']

    def __init__(self, input : any = None, output : any = None) -> None:
        self._input = input
        self._output = output

    async def open(self) -> None:
        raise NotImplementedError
    
    async def read(self, n_bytes : int) -> bytes:
        raise NotImplementedError
        
    async def write(self, byte_string : bytes) -> None:
        raise NotImplementedError
    
    async def close(self) -> None:
        raise NotImplementedError

class file_interface(base_IO_interface):
    '''
        A minimal implemenation of a file interface. Receives an input
        file name and (optionally) an output file name, to which it appends.
    '''
    __slots__ = ['_input', '_output', '_o', '_i']

    def __init__(self, input : any = None, output : any = None) -> None:
        self._input = input
        self._output = output

    async def open(self) -> None:
        self._o = open(self._output, 'ab') if self._output is not None else None
        self._i = open(self._input, 'rb')
    
    async def read(self, n_bytes : int) -> bytes:
        r = self._i.read(n_bytes)
        if r == b'':
            raise EOFError('End of File reached')
        return r
        
    async def write(self, byte_string : bytes) -> None:
        if self._o is not None:
            self._o.write(byte_string)
    
    async def close(self) -> None:
        if self._o is not None:
            self._o.close()
        self._i.close()

class tcp_interface(base_IO_interface):
    '''
        A minimal implementation of a TPC interface. It wraps a
        connection established with the asyncio module.
    '''
    __slots__ = ['_ip', '_port', '_reader', '_writer']

    def __init__(self, ip : str, port : int) -> None:
        self._ip = ip
        self._port = port

    async def open(self) -> None:
        self._reader, self._writer = await _asyncio.open_connection(self._ip, self._port)
    
    async def read(self, n_bytes : int) -> bytes:
        r = await self._reader.read(n_bytes)
        if r == b'':
            raise EOFError('Connection returned empty byte string')
        return r
        
    async def write(self, byte_string : bytes) -> None:
        self._writer.write(byte_string)
        await self._writer.drain()
    
    async def close(self) -> None:
        self._writer.close()
        await self._writer.wait_closed()



import pyimclsts.network as n
import pyimc_generated as pg

import pandas as pd
import numpy as np

import xarray as xra
import csv

csv_delimiter = '\x01'#'; '
json_delimiter = ', '

# \copy lauvxplore1 from 'output.csv' delimiter E'\x01';

def tolist(msg_or_value) -> list:
    if isinstance(msg_or_value, pg._base_templates.base_message):
        fields = msg_or_value.Attributes.fields
        values = [tolist(getattr(msg_or_value, f)) for f in fields if tolist(getattr(msg_or_value, f)) is not None]
        
        return [[f, v] for (f, v) in zip(fields, values)]
    elif isinstance(msg_or_value, list):
        return [tolist(i) for i in msg_or_value]
    elif isinstance(msg_or_value, int) or isinstance(msg_or_value, float):
        return msg_or_value
    else: # ignore plaintext and rawdata
        return None
        
class table():
    
    def __init__(self, f : str) -> None:
        '''f is file name'''

        self.file_name = f
        # erases the file if it exists, creates it otherwise
        with open(f, 'w'):
            pass

        self.datatable = []
        self.estimated_states = []
            
    def writetotable(self, msg, callback) -> str:
        
        time = msg._header.timestamp
        message_abbrev = msg.Attributes.abbrev
        src = msg._header.src
        src_ent = msg._header.src_ent
        
        data = tolist(msg)
        data = [[time, message_abbrev, src, src_ent, *d] for d in data] # i don't think it expects a list with more than 1 item.
        self.datatable += data

    def update_state(self, msg, callback):
        time = msg._header.timestamp
            
        point = [msg.lat, msg.lon, msg.depth, time]
        
        self.estimated_states.append(point)

if __name__ == '__main__':
    
    w = table('output.csv')

    src_file = 'Data.lsf'
    sub = n.subscriber(n.file_interface(input = src_file), use_mp=True)

    sub.subscribe_async(w.writetotable, msg_id =pg.messages.Temperature, src='lauv-noptilus-1', src_ent='AHRS')
    sub.subscribe_async(w.update_state, msg_id =pg.messages.EstimatedState, src='lauv-noptilus-1', src_ent=None)

    sub.run()

    positions = pd.DataFrame(w.estimated_states, columns=['lat', 'lon', 'depth', 'timestamp'])
    values = pd.DataFrame(w.datatable, columns=['timestamp', 'message', 'src', 'src_ent','field', 'value'])

    interpolator = lambda x, key : np.interp(x, positions['timestamp'], positions[key])

    keys = ['lat', 'lon', 'depth']
    for k in keys:
        values[k] = interpolator(values['timestamp'], k)

    metadata = pd.read_csv('AUV_General_Metadata.csv', delimiter=';')
    metadata = metadata.groupby(by = 'Variable Name')

    netcdf_DS = xra.Dataset.from_dataframe(values)


'''
    Contains classes that define the role this pyimc instance will perform and
    are intended to be used directly by the user or their abstraction layer

'''
import typing as _typing
import functools as _functools
import inspect as _inspect

import multiprocessing as _multiprocessing
import asyncio as _asyncio
import time as _time

import importlib.util as _import
import sys as _sys
import os as _os

import pyimclsts.core as _core

_module_name = 'pyimc_generated' # and folder name
_location = _os.getcwd() + '/' + _module_name + '/__init__.py'

_spec = _import.spec_from_file_location(_module_name, _location)

_pg = _import.module_from_spec(_spec)
_sys.modules[_module_name] = _pg
_spec.loader.exec_module(_pg)

def unpack(message : bytes, *, is_big_endian : bool = None, is_field_message : bool = False, fast_mode : bool = False) -> _pg._base_templates.IMC_message:
    '''Expects a serializable (= exactly long (header + fields + CRC)) string of bits whose CRC has already been checked
    
    Fast mode skips all type checking performed by the descriptor by directly invoking the constructor.
    '''
    if is_big_endian is None:
        is_big_endian = int.from_bytes(message[:2], byteorder='big') == _pg._base_templates._sync_number
        # Note: is_big_endian is a function parameter to enable recursion
    
    unpack_functions = _core.unpack_functions_big if is_big_endian else _core.unpack_functions_little
    cursor = 0
    
    if not is_field_message:
        # deserialize header
        (m, size) = unpack_functions['header'](message[cursor:])
        deserialized_header = _pg._base_templates.header_data(*m)
        cursor += size
    
        msgid = deserialized_header.mgid
        if msgid not in _pg.messages._message_ids:
            unknown_msg = _pg.messages.Unknown(msgid, contents = message[cursor:-2], endianess = is_big_endian)
            unknown_msg._header = deserialized_header
            return unknown_msg
    else:
        msgid = unpack_functions['uint16_t'](message[cursor:cursor+2])[0]
        cursor += 2
        if msgid not in _pg.messages._message_ids:
            raise KeyError(f'Cannot parse/unpack an unknown inlined message (no information about the size). Add message id {msgid} to extract list')
    
    if fast_mode:
        # get corresponding class
        message_class = getattr(_pg.messages, _pg.messages._message_ids.get(msgid, None))

        fields = [(f, getattr(getattr(type(message_class()), f), '_field_def')['type']) for f in message_class()._Attributes.fields]
        arguments = dict()
        for field, t in fields:
            if t == 'message':
                if unpack_functions['uint16_t'](message[cursor:cursor+2])[0] == 65535:
                    cursor += 2
                else:
                    (m, size) = unpack(message[cursor:], is_big_endian=is_big_endian, is_field_message=True, fast_mode=fast_mode)
                    arguments[field] = m
                    cursor += size
            elif t == 'message-list':
                (n, _) = unpack_functions['uint16_t'](message[cursor:])
                cursor += 2
                arguments[field] = []
                for _ in range(n):
                    (m, size) = unpack(message[cursor:], is_big_endian=is_big_endian, is_field_message=True, fast_mode=fast_mode)
                    arguments[field].append(m)
                    cursor += size
            else:
                (m, size) = unpack_functions[t](message[cursor:])
                arguments[field] = m
                cursor += size
        # instantiate class through constructor
        message_class = message_class(**arguments)
        
    else:
        # instantiate empty class
        message_class = getattr(_pg.messages, _pg.messages._message_ids.get(msgid, None))()

        # deserialize fields
        # make a (field, type) tuple list, get information in the descriptor
        fields = [(f, getattr(getattr(type(message_class), f), '_field_def')['type']) for f in message_class._Attributes.fields]
        for field, t in fields:
            if t == 'message':
                if unpack_functions['uint16_t'](message[cursor:cursor+2])[0] == 65535:
                    cursor += 2
                else:
                    (m, size) = unpack(message[cursor:], is_big_endian=is_big_endian, is_field_message=True, fast_mode=fast_mode)
                    cursor += size
                    setattr(message_class, field, m)
            elif t == 'message-list':
                (n, _) = unpack_functions['uint16_t'](message[cursor:])
                cursor += 2
                message_list = []
                for _ in range(n):
                    (m, size) = unpack(message[cursor:], is_big_endian=is_big_endian, is_field_message=True, fast_mode=fast_mode)
                    message_list.append(m)
                    cursor += size
                setattr(message_class, field, message_list)
            else:
                (m, size) = unpack_functions[t](message[cursor:])
                cursor += size
                setattr(message_class, field, m)
    
    if not is_field_message:
        message_class._header = deserialized_header
        return message_class
    else:
        return (message_class, cursor)

def _get_id_src_src_ent(message : bytes) -> int:
    src_ent = message[16]
    if int.from_bytes(message[:2], byteorder='big') == _pg._base_templates._sync_number:
        id = int.from_bytes(message[2:4], byteorder='big')
        src = int.from_bytes(message[14:16], byteorder='big')
        
        return (id, src, src_ent)
    else:
        id = int.from_bytes(message[2:4], byteorder='little')
        src = int.from_bytes(message[14:16], byteorder='little')
        
        return (id, src, src_ent)

# Re-export some classes:

tcp_interface = _core.tcp_interface
file_interface = _core.file_interface

class _message_bus():
    '''Injected dependency to 'simplify' common functionalities'''
    __slots__ = ['_io_interface', '_timeout', '_big_endian', '_block_outgoing']

    def __init__(self, IO_interface : _core.base_IO_interface, timeout = 60, big_endian=False):
        self._io_interface = IO_interface
        self._timeout = timeout

        # mode to send messages
        self._big_endian = big_endian
        self._block_outgoing = False
    
    def __enter__(self):
        raise NotImplemented
    def __exit__(self, exc_type, exc_value, exc_tb):
        raise NotImplemented
    
    def block_outgoing(self):
        '''Blocks (and discards) outgoing messages'''
        self._block_outgoing = True
    def unblock_outgoing(self):
        '''Unblock outgoing messages'''
        self._block_outgoing = False

    def send(self, message : _pg._base_templates.base_message, *, src : int = None, src_ent : int = None, 
                        dst : int = None, dst_ent : int = None) -> None:
        '''Wrapper around a queue (actually a pipe end).'''
        if not self._block_outgoing:
            self._send(message, src = src, src_ent = src_ent, dst = dst, dst_ent = dst_ent)
            
    def _send(self, message : _pg._base_templates.base_message, *, src : int = None, src_ent : int = None, 
                        dst : int = None, dst_ent : int = None) -> None:
        raise NotImplemented

class message_bus(_message_bus):
    '''
        Send and receives messages as bytes, but exposes them as IMC messages

        Receives a base_IO_interface, which must implement open(), read(), write() and close()
        asynchronous methods.

        Starts another process that continuously reads/writes to the base_IO_interface.
    '''

    __slots__ = ['_child_end', '_parent_end', '_child_process', '_keep_running', '_big_endian']

    def _external_listener_loop(self, child_end, timeout : int, keep_running : _multiprocessing.Value):
        '''All code bellow is executed in a separate process.'''

        async def consume_output(io_interface : _core.base_IO_interface):
            '''Continuously read the pipe end to send messages'''
            
            while keep_running.value:
                try:
                    has_message = child_end.poll()
                    while has_message:
                        message = child_end.recv_bytes()
                        await io_interface.write(message)
                        has_message = child_end.poll()
                finally:
                    pass
                # Yield an exit point to the event loop
                await _asyncio.sleep(0)
            
            print("Writer stream has been closed.")

        async def consume_input(io_interface : _core.base_IO_interface):
            '''Continuously read the socket to deserialize messages'''

            buffer = bytearray()
            while keep_running.value:
                try:
                    # magic number: 6 = sync number + (msgid + msgsize) size in bytes
                    if len(buffer) < 6:
                        buffer += await io_interface.read(6 - len(buffer))

                    if int.from_bytes(buffer[:2], byteorder='little') == _pg._base_templates._sync_number:
                        # get msg size
                        size = int.from_bytes(buffer[4:6], byteorder='little')
                        # magic number: 22 = 20(header size) + 2(CRC) sizes in bytes.
                        read_size = max(size + 22 - len(buffer), 0)
                        buffer += await io_interface.read(read_size)

                        # Validate message, but do not unpack yet
                        unparsed_msg = bytes(buffer[:(size + 22)])
                        if _core.CRC16IMB(unparsed_msg[:-2]) == int.from_bytes(unparsed_msg[-2:], byteorder='little'):
                            child_end.send_bytes(unparsed_msg)
                            # eliminate message from buffer
                            del buffer[:size + 22]
                        else:
                            # deserialization failed:
                            # sync number is not followed by a sound/valid message. Remove it from buffer
                            # to look for next message
                            del buffer[:2]
                    elif int.from_bytes(buffer[:2], byteorder='big') == _pg._base_templates._sync_number:
                        size = int.from_bytes(buffer[4:6], byteorder='big')
                        read_size = max(size + 22 - len(buffer), 0)
                        buffer += await io_interface.read(read_size)

                        unparsed_msg = bytes(buffer[:(size + 22)])
                        if _core.CRC16IMB(unparsed_msg[:-2]) == int.from_bytes(unparsed_msg[-2:], byteorder='big'):
                            child_end.send_bytes(unparsed_msg)
                            del buffer[:size + 22]
                        else:
                            del buffer[:2]
                    else:
                        # buffer does not start with a sync number. Remove it to search
                        # for a valid sync number.
                        del buffer[:2]
                except EOFError as e:
                    print("EOF reached by the stream reader. Waiting for stream writer to finish...")
                    
                    # Unblock the main thread and send an empty byte string. 
                    # (-> signal EOF, so that it won't write anymore)
                    child_end.send_bytes(b'')

                    # Yield to the event loop to let stream writer finish 
                    await _asyncio.sleep(1.5)
                    
                    # Prevent further reads/writes in this process
                    with keep_running.get_lock():
                        keep_running.value = False
                finally:
                    pass
                
                # Yield an exit point to the event loop
                await _asyncio.sleep(0)
            print("Reader stream has been closed.")
        async def main_loop():
            await self._io_interface.open()
            try:
                await _asyncio.gather(consume_input(self._io_interface), consume_output(self._io_interface))
            finally:
                child_end.close()
                print('IO interface has been closed.')
                await self._io_interface.close()

        try:
            _asyncio.run(main_loop())
        except EOFError as e:
            print('No more bytes to read.')
        finally:
            with self._keep_running.get_lock():
                self._keep_running.value = False
            print('Message Bus has been closed.')

    def open(self):
        
        # Using a pipe to establish communication between processes
        self._parent_end, self._child_end = _multiprocessing.Pipe(duplex=True)

        self._keep_running = _multiprocessing.Value('i', True)

        # Start process
        self._child_process = _multiprocessing.Process(target=self._external_listener_loop, 
                                                        args=(self._child_end, self._timeout, self._keep_running))
        self._child_process.start()

        # It is very likely that the main process will run faster than the child process, which
        # may cause some undesirable behaviour, such as, the main process' context manager closes 
        # the connection before the child process' procedures can even start.
        # The naive solution: block the main thread for 0.5 second
        _time.sleep(0.5)
    
    def close(self, max_wait : float = 1) -> None:        
        with self._keep_running.get_lock():
            self._keep_running.value = False
        
        self._child_process.join()
        self._child_process.close()

    def _send(self, message : _pg._base_templates.base_message, *, src : int = None, src_ent : int = None, 
                        dst : int = None, dst_ent : int = None) -> None:
        self._parent_end.send_bytes(message.pack(is_big_endian=self._big_endian, src = src, src_ent = src_ent, 
                        dst = dst, dst_ent = dst_ent))

    def recv(self) -> _pg._base_templates.base_message:
        '''Wrapper around a queue (actually a pipe end). Blocks until a message is available.
        The _external_listener_loop is supposed to send complete messages (as per multiprocessing 
        documentation).'''       
        msg = self._parent_end.recv_bytes()
        
        if msg == b'':
            raise EOFError('Message Bus has been closed.')
        else:
            return msg
            
    def poll(self, timeout : int = 0) -> bool:
        '''Extra function to check whether there are any available messages.
        Check _multiprocessing module pipes.
        '''
        return self._parent_end.poll(timeout=timeout)

    def __enter__(self):
        self.open()
        return self
    
    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()
        if exc_type == EOFError:
            print('Child process has been closed due to end of file.')
            return True
        print('Child process has been closed.')
        return None

class message_bus_st(_message_bus):
    '''
        Send and receives messages as bytes, but exposes them as IMC messages

        Receives a base_IO_interface, which must implement open(), read(), write() and close()
        asynchronous methods.

        DOES NOT start another process. Runs in the main process.
    '''

    __slots__ = ['_writer_queue', '_reader_queue', '_keep_running', '_big_endian', '_task']
    
    async def open(self):
        self._keep_running = True

        self._writer_queue = _asyncio.Queue()
        self._reader_queue = _asyncio.Queue()

        async def consume_output(io_interface : _core.base_IO_interface):
            '''Continuously read the pipe end to send messages'''
            
            while self._keep_running:
                try:
                    # flush queue or wait.
                    while not self._writer_queue.empty():
                        message = await self._writer_queue.get()
                        await io_interface.write(message)
                    else:
                        await _asyncio.sleep(0.5)
                finally:
                    pass
            
            print("Writer stream has been closed.")

        async def consume_input(io_interface : _core.base_IO_interface):
            '''Continuously read the socket to deserialize messages'''
            
            buffer = bytearray()
            while self._keep_running:
                try:
                    # magic number: 6 = sync number + (msgid + msgsize) size in bytes
                    if len(buffer) < 6:
                        buffer += await io_interface.read(6 - len(buffer))

                    if int.from_bytes(buffer[:2], byteorder='little') == _pg._base_templates._sync_number:
                        # get msg size
                        size = int.from_bytes(buffer[4:6], byteorder='little')
                        # magic number: 22 = 20(header size) + 2(CRC) sizes in bytes.
                        read_size = max(size + 22 - len(buffer), 0)
                        buffer += await io_interface.read(read_size)

                        # Validate message, but do not unpack yet
                        unparsed_msg = bytes(buffer[:(size + 22)])
                        if _core.CRC16IMB(unparsed_msg[:-2]) == int.from_bytes(unparsed_msg[-2:], byteorder='little'):
                            await self._reader_queue.put(unparsed_msg)
                            # eliminate message from buffer
                            del buffer[:size + 22]
                        else:
                            # deserialization failed:
                            # sync number is not followed by a sound/valid message. Remove it from buffer
                            # to look for next message
                            del buffer[:2]
                    elif int.from_bytes(buffer[:2], byteorder='big') == _pg._base_templates._sync_number:
                        size = int.from_bytes(buffer[4:6], byteorder='big')
                        read_size = max(size + 22 - len(buffer), 0)
                        buffer += await io_interface.read(read_size)

                        unparsed_msg = bytes(buffer[:(size + 22)])
                        if _core.CRC16IMB(unparsed_msg[:-2]) == int.from_bytes(unparsed_msg[-2:], byteorder='big'):
                            await self._reader_queue.put(unparsed_msg)
                            del buffer[:size + 22]
                        else:
                            del buffer[:2]
                    else:
                        # buffer does not start with a sync number. Remove it to search
                        # for a valid sync number.
                        del buffer[:2]
                except EOFError as e:
                    print("EOF reached by the stream reader. Waiting for stream writer to finish...")
                    
                    # Unblock the main thread and send an empty byte string. 
                    # (-> signal EOF, so that it won't write anymore)
                    await self._reader_queue.put(b'')

                    # Yield to the event loop to let stream writer finish 
                    await _asyncio.sleep(1.5)
                    
                    # Prevent further reads/writes in this process
                    self._keep_running = False
                finally:
                    pass
                await _asyncio.sleep(0)
            print("Reader stream has been closed.")
        async def main_loop():
            await self._io_interface.open()
            try:
                await _asyncio.gather(consume_input(self._io_interface), consume_output(self._io_interface))
            finally:
                print('IO interface has been closed.')
                await self._io_interface.close()

        self._task = _asyncio.create_task(main_loop())
    
    def close(self, max_wait : float = 1) -> None:
        self._keep_running = False
        self._task.cancel()
        print('Message Bus has been closed.')

    def _send(self, message : _pg._base_templates.base_message, *, src : int = None, src_ent : int = None, 
                        dst : int = None, dst_ent : int = None) -> None:
        self._writer_queue.put_nowait(message.pack(is_big_endian=self._big_endian, src = src, src_ent = src_ent, 
                        dst = dst, dst_ent = dst_ent))

    async def recv(self) -> _pg._base_templates.base_message:
        '''Wrapper around a queue (actually a pipe end). Blocks until a message is available.
        The _external_listener_loop is supposed to send complete messages (as per multiprocessing 
        documentation).'''

        msg = await self._reader_queue.get()
        
        if msg == b'':
            raise EOFError('No more bytes to read.')
        else:
            return msg
            
    def poll(self, timeout : int = 0) -> bool:
        '''Extra function to check whether there are any available messages.
        '''
        return not self._reader_queue.empty()
    
    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()
        if exc_type == EOFError:
            print('Message bus event loop has been closed due to end of file.')
            return True
        print('Message bus event loop has been closed.')
        return None

class subscriber:

    __slots__ = ['_msg_manager', '_subscriptions', '_subscripted_all', '_periodic', '_call_once', '_use_mp', '_peers', '_src2name']

    def __init__(self, IO_interface : _core.base_IO_interface, *,big_endian=False, use_mp = False) -> None:
        self._use_mp = use_mp
        if self._use_mp:
            self._msg_manager = message_bus(IO_interface, big_endian)
        else:
            self._msg_manager = message_bus_st(IO_interface, big_endian)
        self._subscriptions = dict()
        self._subscripted_all = []
        self._periodic = []
        self._call_once = []

        # a dictionary of {vehicle name : {'src' : 1, 'entities' : { 1 : 'Entity name'...} ...}}
        self._peers = dict()
        # a dictionary of {2: 'vehicle name'}
        self._src2name = dict()
        
        self.subscribe_async(self._abort, _pg.messages.Abort)
        
        self.call_once(self._queryEntityList, delay=1)
        self.periodic_async(self._queryEntityList, period=300)
        self.subscribe_async(self._update_peers, _pg.messages.EntityList)
        self.subscribe_async(self._update_peers, _pg.messages.Announce)

    async def _periodic_wrapper_coro(self, _period : float, f : _typing.Callable, send_callback : _typing.Callable[[_pg._base_templates.IMC_message], None]):
        loop = _asyncio.get_running_loop()
        while True:
            last_exec = loop.time()
            await f(send_callback)
            now = loop.time()
            await _asyncio.sleep(max(last_exec - now + _period, 0))
                
    async def _periodic_wrapper(self, _period : float, f : _typing.Callable, send_callback : _typing.Callable[[_pg._base_templates.IMC_message], None]):
        loop = _asyncio.get_running_loop()
        f(send_callback)
        while True:
            await _asyncio.sleep(_period)
            loop.call_later(_period, f, send_callback)

    async def _event_loop(self):
        msg_mgr = self._msg_manager
        try:
            if self._use_mp:
                msg_mgr.open()
            else:
                await msg_mgr.open()
            loop = _asyncio.get_running_loop()

            for f, delay in self._call_once:
                if delay is not None:
                    loop.call_later(delay, f, msg_mgr.send)
                else:
                    f(msg_mgr.send)
            
            # Binding references to background task objects so that they are not garbage collected
            tasks = []
            for f, period in self._periodic:
                if _inspect.iscoroutinefunction(f):
                    tasks.append(loop.create_task(self._periodic_wrapper_coro(period, f, msg_mgr.send)))
                elif callable(f):
                    tasks.append(loop.create_task(self._periodic_wrapper(period, f, msg_mgr.send)))
                else:
                    print(f'Warning: Given function {f} is neither _typing.Callable nor a coroutine.')

            while True:
                msg = msg_mgr.recv() if self._use_mp else await msg_mgr.recv()
                mgid, src, src_ent = _get_id_src_src_ent(msg)
                if mgid in self._subscriptions:
                    desel_message = unpack(msg, fast_mode=True)
                    for f in self._subscriptions[mgid]:
                        if self._validate_call(src, src_ent, f[1], f[2]):
                            await f[0](desel_message, msg_mgr.send)
                
                [await f[0](unpack(msg, fast_mode=True), msg_mgr.send) for f in self._subscripted_all if self._validate_call(src, src_ent, f[1], f[2])]
                # Offer an exit point
                await _asyncio.sleep(0)
        except EOFError:
            print('Stream has ended.')
        finally:
            msg_mgr.close()

    async def _abort(self, msg, send_callback):
        if msg._header is not None:
            my_src = 0x4000 | (_core.get_initial_IP() & 0xFFFF)
            if msg._header.dst == my_src:
                loop = _asyncio.get_running_loop()
                loop.close()

    def _update_peers(self, msg : _typing.Union[_pg.messages.EntityList, _pg.messages.Announce], send_callback):
        if msg._header is not None:
            src = msg._header.src
            
            if isinstance(msg, _pg.messages.EntityList):
                if msg.op == msg.OP.REPORT:
                    entList =  [i.split(sep='=') for i in msg.list.split(sep=';')]
                    entList = {k : int(v) for [k, v] in entList}

                    name = self._src2name.get(src, None)
                    if name is not None:
                        # if it exists, update; else, create entry
                        if self._peers.get(name, None) is not None:
                            self._peers[name]['EntityList'] = entList
                        else:
                            self._peers[name] = {'EntityList' : entList}

            elif isinstance(msg, _pg.messages.Announce):
                name = msg.sys_name

                self._src2name[src] = name
                # if it exists, update; else, create entry
                if self._peers.get(name, None) is not None:
                    self._peers[name]['src'] = src
                else:
                    self._peers[name] = {'src' : src}
        else:
            pass
    
    def _get_src(self, vehicle_name : str):
        return self._peers[vehicle_name].get('src', None) if self._peers.get(vehicle_name, None) is not None else None
    
    def _get_src_ent(self, vehicle_name : str, entityName : str):
        entityList = self._peers[vehicle_name].get('EntityList', None) if self._peers.get(vehicle_name, None) is not None else None
        if entityList is not None:
            return entityList.get(entityName, None)
        return None
    
    def _queryEntityList(self, send):
        query = _pg.messages.EntityList(op=_pg.messages.EntityList.OP.QUERY, list='')
        send(query)
    
    def _validate_call(self, src, src_ent, desired_src : str, desired_src_ent : str) -> bool:
        if desired_src is None and desired_src_ent is None:
            return True
        else:
            correct_src = src == self._get_src(desired_src) 
            correct_src_ent = src_ent == self._get_src_ent(desired_src, desired_src_ent)
            
            if (correct_src or correct_src_ent) and (desired_src is None or desired_src_ent is None):
                return True
            elif correct_src and correct_src_ent:
                return True
            
        return False
    
    def subscribe_async(self, callback : _typing.Callable[[_pg._base_templates.IMC_message], None], msg_id : _typing.Union[int, _pg._base_templates.IMC_message] = None, *, src : str = None, src_ent : str = None):
        '''Appends the callback to the list of subscripted functions.
        msg_id can be provided as an int or the class of the message.
        src and src_ent can be provided as strings. If None, then defaults to all.
        
        The main loop calls the function and pass the message and a callback to send messages.
        The return value is discarded and the function must have exactly two parameters (the message and the callback)
        and must not have additional parameters, to avoid unintended behavior*.
        
        *Due to python's pass-by-assignment, the argument values are evaluated when a function is called.
        This means that even if we pass, say "x", to the .subscribe_async method and then try to
        pass "x" to the callback, we can only pass x's value at the moment .subscribe_async was called.
        This means that if we subscribe, for example, f(x) and somewhere a function g "updates" x, and then call
        f(x) again, f will be called with x's first value.
        This could be bypassed using mutable data structures, but let's not delve into that.

        Tip: If the original function really needs arguments, wrap it with _functools.partial.
        Tip2: Use a class instance to keep shared values across different calls. See followRef.py.
        '''
        key = None
        if isinstance(msg_id, _pg._base_templates.IMC_message):
            key = msg_id._Attributes.id
        elif isinstance(msg_id, int):
            key = msg_id
        elif _inspect.isclass(msg_id):
            if issubclass(msg_id, _pg._base_templates.IMC_message):
                key = msg_id()._Attributes.id
        elif msg_id is None:
            pass
        else:
            print(f'Warning: Given message id {msg_id} is not a known message.')
            
        # There is either an explicit subscription (the key is not None) or 'all'
        # (msg_id is None). There cannot be both nor neither.
        if (key is not None) ^ (msg_id is None):
            c = None
            if _inspect.iscoroutinefunction(callback):
                c = callback
            elif callable(callback):
                c = _functools.partial(_core._async_wrapper, callback)
            else:
                print(f'Warning: Given function {callback} is neither callable nor a coroutine.')
            
            if c is not None:
                if msg_id is None:
                    self._subscripted_all.append((c, src, src_ent))
                else:
                    if self._subscriptions.get(key, None) is not None:
                        self._subscriptions[key].append((c, src, src_ent))
                    else:
                        self._subscriptions[key] = [(c, src, src_ent)]

    def periodic_async(self, callback : _typing.Callable[[_pg._base_templates.IMC_message], None], period = float):
        '''Add callback to a list to be called every period seconds. Function must take
        a send callback as parameter. This callback can be used to send messages.'''
        self._periodic.append((callback, period))

    def subscribe_mp(self, callback : _typing.Callable[[_pg._base_templates.IMC_message], None], msg_id : _typing.Union[int, _pg._base_templates.IMC_message], *, src : str = None, scr_ent : str = None):
        '''Calls a function and pass the message and a callback to send messages to it.
        Runs the given callback in a different process and should be used only with heavy load
        functions.
        '''
        raise NotImplemented

    def call_once(self, callback : _typing.Callable[[_typing.Callable[[_pg._base_templates.IMC_message], None]], None], delay : float = None):
        '''Calls the given callbacks as soon as the main loop starts or according to their delay in seconds.
        Callback parameters must be exactly one callback (that can be used to send messages).'''
        self._call_once.append((callback, delay))

    def run(self):
        _asyncio.run(self._event_loop())


import pyimclsts.network as p
import pyimc_generated as pg
import random 

class FollowRef_Vehicle():
    '''
    Minimal implementation to start a Follow Reference manuever
    '''
    __slots__ = ['EstimatedState', 'FollowRefState', 'peers', 'target', 'request_id']
    
    def __init__(self, target : str) -> None:
        '''target is the name of the vehicle as in Announce messages'''
        self.EstimatedState = None
        self.FollowRefState = None
        
        self.target = target
        self.peers = dict()

    def request_followRef(self, send_callback):
        if self.FollowRefState is None:
            request = pg.messages.PlanControl()
            
            fr = pg.messages.FollowReference()
            fr.control_src = 0xFFFF #0x4000 | (pg.core.get_initial_IP() & 0xFFFF)
            fr.control_ent = 0xFF
            fr.timeout = 10
            fr.loiter_radius = 0
            fr.altitude_interval = 0

            request.type = pg.messages.PlanControl.TYPE.REQUEST
            request.op = pg.messages.PlanControl.OP.START
            self.request_id = random.randint(0, 0xFFFF)
            request.request_id = self.request_id
            request.plan_id = "MyPlan-pyimctrans"
            request.flags = pg.messages.PlanControl.FLAGS.IGNORE_ERRORS
            request.arg = fr
            request.info = "MyPlan"
            
            print("Requesting Follow Reference Manuever...")
            print(request)
            send_callback(request, dst = self.peers.get(self.target, 0xFFFF))

    def send_refs(self, send_callback):
        if self.EstimatedState is not None:
            referenceToFollow = pg.messages.Reference()

            referenceToFollow.flags = pg.messages.Reference.FLAGS.LOCATION | pg.messages.Reference.FLAGS.SPEED | pg.messages.Reference.FLAGS.Z
            referenceToFollow.speed = pg.messages.DesiredSpeed(value=50, speed_units=pg.enumerations.SpeedUnits.PERCENTAGE)
            referenceToFollow.z = pg.messages.DesiredZ(value=1, z_units=pg.enumerations.ZUnits.DEPTH)
            referenceToFollow.lat = self.EstimatedState.lat + 0.001
            referenceToFollow.lon = self.EstimatedState.lon
            referenceToFollow.radius = 0

            send_callback(referenceToFollow, dst = self.peers.get(self.target, 0xFFFF))

    def update_vehicle_state(self, msg : pg.messages.EstimatedState, send_callback):
        self.EstimatedState = msg

    def update_plan_state(self, msg : pg.messages.FollowRefState, send_callback):
        self.FollowRefState = msg
    
    def update_peers(self, msg : pg.messages.Announce, send_callback):
        self.peers[msg.sys_name] = msg._header.src

if __name__ == '__main__':
    
    conn = p.tcp_interface('localhost', 6006)
    vehicle = FollowRef_Vehicle('lauv-xplore-2')

    sub = p.subscriber(conn)

    # Set a delay, so that we receive the Announcements
    sub.call_once(vehicle.request_followRef, 5)

    sub.subscribe_async(vehicle.update_peers, pg.messages.Announce)
    sub.subscribe_async(vehicle.update_vehicle_state, pg.messages.EstimatedState)
    sub.subscribe_async(vehicle.update_plan_state, pg.messages.FollowReference)

    sub.periodic_async(vehicle.send_refs, 1)

    #sub.periodic_async(lambda _ : print(vehicle.peers), 1)

    sub.run()


import pyimclsts.network as n
import pyimc_generated as pg

import base64
import json

csv_delimiter = '\x01'#'; '
json_delimiter = ', '

# \copy lauvxplore1 from 'output.csv' delimiter E'\x01';

# Jesus Christ, it's become so difficult/ugly to program outside of Haskell...
def tojson(msg_or_value) -> str:
    if isinstance(msg_or_value, pg._base_templates.base_message):
        fields = msg_or_value.Attributes.fields
        values = [tojson(getattr(msg_or_value, f)) for f in fields]
        
        msg_id = f'"Message id" : {msg_or_value.Attributes.id}'

        json_string = json_delimiter.join([msg_id] + [f'"{field}" : {value}' for field, value in zip(fields, values)])
        
        return '{' + json_string + '}' 
    elif isinstance(msg_or_value, list):
        return '[' + json_delimiter.join([tojson(i) for i in msg_or_value]) + ']'
    elif isinstance(msg_or_value, int) or isinstance(msg_or_value, float):
        return str(msg_or_value)
    elif isinstance(msg_or_value, bytes):
        if msg_or_value != b'':
            return f'"{str(base64.b64encode(msg_or_value), encoding="ascii")}"'
        else:
            return 'null'
    else: # plaintext. Remove single quotation marks ('), re-escape escape characters, escape (")
        if msg_or_value:
            #x = repr(json.dumps(msg_or_value.encode("ascii", errors="surrogateescape").decode())[1:-1])[1:-1]
            #return f'"{x}"'
            # convert plaintext to its binary representation, then to its base64 rep.
            # Too much of a hassle to escape characters and deal with invalid ASCII.
            return f'"{str(base64.b64encode(msg_or_value.encode("ascii", errors="surrogateescape")), encoding="ascii")}"'
        else:
            return "null"

def todict(msg_or_value) -> dict:
    if isinstance(msg_or_value, pg._base_templates.base_message):
        fields = list(msg_or_value.Attributes.fields)
        values = [todict(getattr(msg_or_value, f)) for f in fields]
        
        json_dict = { f : v for f, v in zip(["Message id"] + fields, [msg_or_value.Attributes.id] + values) }
        
        return json_dict
    elif isinstance(msg_or_value, list):
        return [todict(i) for i in msg_or_value]
    elif isinstance(msg_or_value, bytes):
        return str(base64.b64encode(msg_or_value), encoding="ascii")
    else:
        return msg_or_value

class writer():
    
    def __init__(self, f : str, b_size : int) -> None:
        '''f is file name, b_size is buffer size in bytes'''

        self.file_name = f
        # erases the file if it exists, creates it otherwise
        with open(f, 'w'):
            pass
        self.buffer_max_size = b_size
        self.buffer = []

        self.buffer_size = 0
    
    def write(self, s : str) -> None:
        self.buffer.append(s)
        self.buffer_size += len(s)

        if self.buffer_size > self.buffer_max_size:
            with open(self.file_name, 'a') as f:
                f.write('\n'.join(self.buffer) + '\n')
                self.buffer = []
                self.buffer_size = 0
            
    def writetocsv(self, msg, callback) -> str:
        # Note: removed sync number. Size kept?
        header = csv_delimiter.join([str(i) for i in list(msg._header)[1:]])
        
        json_string = tojson(msg)
        self.write(header + csv_delimiter + json_string)
    
    def flush(self):
        with open(self.file_name, 'a') as f:
            f.write('\n'.join(self.buffer))
            self.buffer = []
            self.buffer_size = 0

class writer2():
    
    def __init__(self, f : str, b_size : int) -> None:
        '''f is file name, b_size is buffer size in bytes'''

        self.file_name = f
        # erases the file if it exists, creates it otherwise
        with open(f, 'w'):
            pass
            
    def writetocsv(self, msg, callback) -> str:
        # Note: removed sync number. Size kept?
        header = csv_delimiter.join([str(i) for i in list(msg._header)[1:]])
        
        with open(self.file_name, 'a') as f:
            f.write(header + csv_delimiter)
            json.dump(todict(msg), f, ensure_ascii = True)
            f.write('\n')

if __name__ == '__main__':
    
    src_file = 'Data.lsf'
    sub = n.subscriber(n.file_interface(input = src_file))

    w = writer('output.csv', 1E6)
    sub.subscribe_all(w.writetocsv)

    sub.run()
    w.flush()


import copy
from enum import IntEnum, IntFlag
from collections import namedtuple
import time

import pyimclsts.core as core
from . import enumerations as imc_enums
from . import bitfields as imc_bitf

'''
Implementation of mutable and immutable types

Considered operations to be protected:
    - Assignment;
        - Attributes should not be re-assigned to invalid types
    - Access (getter);
        - Reference leakage (when a reference to a mutable object is inadvertently exposed)
    - Initialization.
        - Attributes should not be initialized with invalid types
    - Dynamic creation of attributes
        - Attributes should not be created/added to the module classes (Even if this is not
        dangerous per se, this might induce programming errors)
'''

# "Global" definitions
header_data = namedtuple('header_data', ['sync', 'mgid', 'size', 'timestamp', 'src', 'src_ent', 'dst', 'dst_ent'])
MessageAttributes = namedtuple('MessageAttributes', %MESSAGE_ATTRIBUTES%)

# "Global" variables
_imc_options = ''
_sync_number = %SYNCH_NUMBER%
_default_src = 0x4000 | (core.get_initial_IP() & 0xFFFF)

class IMC_message():
    
    __slots__ = []
    
    '''Empty message to break circular reference.

    Initially, there was 'base_message' and the descriptor 'mutable_attr', which contained a reference
    to a dictionary 'imc_types' that pointed back to base message. This is/was necessary to allow type 
    checking a message whose field may contain other messages, that is, the message class must have a
    descriptor object (as per protocol) and to check if it is of type 'message', it must have a reference 
    to the 'message' class. However, this leads to the cyclic reference:
            
            'base_message' -> descriptor -> dictionary -> 'base_message'.
    
    Although python may be able to handle circular references, we choose not to.

    Therefore, to circumvent this issue, an empty message is added to break this cycle.
            'IMC_message' <- base_message -> descriptor
            'IMC_message' <- dictionary   <- descriptor
    
    Because a parent class does not point to its child classes, there is no cycle.
    '''
    pass

imc_types = %IMC_TYPES%

class base_message(IMC_message):
    
    __slots__ = ['_header', '_footer', '_Attributes']

    def __str__(self) -> str:
        output = ['Message \'' + self._Attributes.name + '\':', 'Fields:']
        for field in self._Attributes.fields:
            value = getattr(self, field)
            if value is not None:
                if type(value) == str:
                    value = '\'' + value + '\''
                
                #Check whether it is an Enumerated type
                elif getattr(getattr(type(self), field), '_field_def').get('unit', None) in ['Enumerated', 'Bitfield']:
                    # check whether the Enum was "validated" by the descriptor
                    if isinstance(value, IntEnum) or isinstance(value, IntFlag):
                        enum_def = getattr(getattr(type(self), field), '_field_def').get('enum-def', None) if getattr(getattr(type(self), field), '_field_def').get('enum-def', None) \
                            else getattr(getattr(type(self), field), '_field_def').get('bitfield-def', None)
                        
                        # check whether it is local or global
                        if not enum_def:
                            value = self._Attributes.abbrev + '.' + str(value)
                        else:
                            value = str(value)
                    else:
                        value = str(value)

                elif isinstance(value, IMC_message):
                    value = ('\n' + str(value)).replace('\n', '\n    ')
                elif type(value) == list:
                    value = ('\n[\n' + '\n'.join([str(v) for v in value]) + '\n]').replace('\n', '\n    ')
                else:
                    value = str(value)
            else:
                value = 'None'

            output.append('  - ' + field + " = " + value)
        output = '\n'.join(output)
        if hasattr(self, '_header'):
            output = repr(self._header) + '\n' + output
        return output

    def __repr__(self) -> str:
        arguments = []
        for field in self._Attributes.fields:
            value = getattr(self, field)            
            if value is not None:
                #Check whether it is an Enumerated type
                if getattr(getattr(type(self), field), '_field_def').get('unit', None) in ['Enumerated', 'Bitfield']:
                    # check whether the Enum was "validated" by the descriptor
                    if isinstance(value, IntEnum) or isinstance(value, IntFlag):
                        enum_def = getattr(getattr(type(self), field), '_field_def').get('enum-def', None) \
                            if getattr(getattr(type(self), field), '_field_def').get('enum-def', None) \
                            else getattr(getattr(type(self), field), '_field_def').get('bitfield-def', None)

                        # check whether it is local or global
                        if enum_def:
                            value = str(value)
                        else:
                            value = self._Attributes.abbrev + '.' + str(value)
                    else:
                        value = repr(value)
                else:
                    value = repr(value)
            else:
                value = 'None'

            arguments.append(field + " = " + value)
        arguments = ', '.join(arguments)
        output = self._Attributes.abbrev + '({})'.format(arguments)
        return output

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, base_message):
            # if both of them have already defined a '_header'
            if hasattr(self, '_header') and hasattr(__o, '_header'):
                if self._header != __o._header:
                    return False
            # if any of them does not have '_header' -> happens when a message is 
            # inside another, I'll skip and check the contents
            
            if hasattr(self, '_header') and hasattr(__o, '_header'):
                if self._header != __o._header:
                    return False
            fields = self._Attributes.fields
            if any([getattr(self, field) != getattr(__o, field) for field in fields]):
                return False
            return True
        return False

    def _pack_fields(self, *, serial_functions : dict) -> bytes:
        # Check if any field is empty (None) and not type 'message' (checked through the descriptor)
        if any([getattr(self, '_' + field) is None for field in self._Attributes.fields if getattr(getattr(type(self), field), '_field_def').get('type', None) != 'message']):
            raise ValueError('Cannot serialize a message that contains an empty (NoneType) field that is not a message.')
        
        serial_functions = serial_functions

        serialized_fields = []
        for field in self._Attributes.fields:
            # Access variable information through the descriptor
            datatype = getattr(getattr(type(self), field), '_field_def').get('type', None)
            
            # check if it is a "NULL" message
            if datatype == 'message' and getattr(self, '_' + field) is None:
                serialized_fields.append(serial_functions['uint16_t'](65535))
            else:
                serialized_fields.append(serial_functions[datatype](getattr(self, '_' + field)))
        
        return b''.join(serialized_fields)

    def _pack_header(self, *, serial_functions : dict, size : int, src : str = None, src_ent : str = None, dst : str = None, dst_ent : str = None) -> bytes:
        '''Gathers necessary builds the header, stores it in the private variable '_header' and returns the bit string.
        
        The tricky part is: Are the header fields fixed or not, that is, they should be hardcoded or not?
        There may be mutability on their: 1. existence; 2, name; 3. order; 4. type (and size).
        
        We will assume that only 4. can change.
            - As a result a namedTuple is globally defined.'''

        mgid = self._Attributes.id

        # If None or a "default" value, overwrite
        if not hasattr(self, '_header'):
            _timestamp = time.time()
            _src = src if src is not None else _default_src
            _src_ent = src_ent if src_ent is not None else 0xFF
            _dst = dst if dst is not None else 0xFFFF
            _dst_ent = dst_ent if dst_ent is not None else 0xFF
        else:
            _timestamp = self._header.timestamp
            _src = src if src is not None else self._header.src
            _src_ent = src_ent if src_ent is not None else self._header.src_ent
            _dst = dst if dst is not None else self._header.dst
            _dst_ent = dst_ent if dst_ent is not None else self._header.dst_ent
        
        header_fields_values = header_data(sync=_sync_number, mgid=mgid, size=size, timestamp=_timestamp, src=_src, src_ent=_src_ent, dst=_dst, dst_ent=_dst_ent)
        
        self._header = header_fields_values

        return serial_functions['header'](*self._header)
    
    def pack(self, *, is_field_message : bool = False, is_big_endian : bool = True, src : int = None, src_ent : int = None, 
                        dst : int = None, dst_ent : int = None) -> bytes:
        '''Serialize function that optionally overwrites the header, if parameters are provided.'''
        
        serial_functions = core.pack_functions_big if is_big_endian else core.pack_functions_little
        
        s_fields = self._pack_fields(serial_functions=serial_functions)
        
        if not is_field_message:
        
            s_header = self._pack_header(serial_functions=serial_functions, size=(len(s_fields)), src=src, src_ent=src_ent, dst=dst, dst_ent=dst_ent)
            s_message = s_header + s_fields
            
            # footer:
            '''Calculates CRC-16 IBM of a bit string'''
            self._footer = core.CRC16IMB(s_message)
            s_message = s_message + serial_functions['uint16_t'](self._footer)

            return s_message
        return serial_functions['uint16_t'](self._Attributes.id) + s_fields

    def get_timestamp(self) -> float:
        '''Get the timestamp. Returns None if the message has no header yet.'''
        if hasattr(self, '_header'):
            return self._header.timestamp
        return None
    
class immutable_attr():
    '''Describes an immutable attribute. The type should be already known at run time, that is,
    included in the class attribute definition of the message (and therefore, it does not need
    to be type checked).
    Additionally, to prevent accidental modification of python mutable types, which may lead to 
    an inconsistent state, it checks whether the attribute is of a immutable type before 
    returning the (reference of) the desired object and if the object is mutable, it returns a 
    deep copy.'''

    def __init__(self, doc : str) -> None:
        self.__doc__ = doc

    def __set_name__(self, owner : any, name : str):
        self._name = '_' + name

    def __get__(self, instance : any, owner : any):
        if instance is None: #some hacky thing to allow docstrings
            return self

        if isinstance(getattr(instance, self._name), (int, float, str, bool, tuple)):
            return getattr(instance, self._name)
        else:
            return copy.deepcopy(getattr(instance, self._name))

    def __set__(self, owner : any, value : any):
        raise AttributeError('Attribute \'{}\' of {} cannot be modified'.format(self._name, type(owner)))
    
    def __delete__(self, __name: str) -> None:
        raise NotImplementedError

class mutable_attr():
    '''Describes a mutable attribute. The type should be already known at run time, that is,
    included in the class attribute definition of the message.
    Additionally, to prevent accidental modification of python mutable types, which may lead to 
    an inconsistent state, it checks whether the attribute is of a immutable type before 
    returning the (reference of) the desired object and if the object is mutable, it returns a 
    deep copy (new and different reference).
    Since it can still be re-assigned, it will be type checked when this operation is carried out.

    Realization: Inside a message: all attributes are immutable, fields attributes are immutable
    fields contents are mutable.
    
    Also, there is no pythonic way of *prohibiting* (ex: type checker) the assignment of a wrong 
    type, which means it *will* fail. So, failure at the descriptor or failure at the packing 
    step... Are failures/exceptions during runtime either way => bad. I'll leave this implementation
    anyway, since this might give better error messages, but it is, all in all, pointless.
    '''

    def __init__(self, field_def : dict, doc : str) -> None:
        self._field_def = field_def
        self.__doc__ = doc

    def __set_name__(self, owner : any, name : str):
        self._priv_name = '_' + name
        self._owner = owner

    def __get__(self, instance : any, owner : any):
        if instance is None: #some hacky thing to allow docstrings
            return self

        # return bare attribute if it is immutable.
        if isinstance(getattr(instance, self._priv_name), (int, float, str, bool, tuple)):
            return getattr(instance, self._priv_name)
        else:
            return copy.deepcopy(getattr(instance, self._priv_name))

    def __set__(self, obj : any, value : any) -> None:
        '''Performs type and boundary checks and throws exceptions'''
        set_value = value
        attribute_type = imc_types.get(self._field_def.get('type', None), None)

        if not attribute_type:
            raise KeyError('Could not find a type declaration for {} in given IMC definition'.format(self._priv_name[1:]))

        # Special and only case of upcasting internally allowed.
        if isinstance(set_value, int) and attribute_type == float:
            set_value = float(set_value)

        if isinstance(set_value, attribute_type):
            '''Obs: Should type casting be implemented? eg.: float -> int
            Type casting/coersion will not be implemented to avoid reinforcing bad
            pratices and increase transparency.
            
            On a 2nd thought, we may allow SOME upcasting, in particular, int -> float'''
            
            # if it is a list, check its elements types.
            if isinstance(set_value, imc_types['message-list']):
                
                for t in set_value:
                    # Check data type
                    if not isinstance(t, imc_types['message']):
                        raise ValueError('Cannot assign {} to attribute \'{}\'. Expected: {} of {}'.format(
                        type(t), self._priv_name[1:], imc_types[self._field_def['type']], 
                        imc_types['message']))
                    
                    # (I have to check the message group?)
                    # Check message-type 
                    if  self._field_def.get('message-type', None) is not None \
                                    and t.Attributes.abbrev != self._field_def.get('message-type', None):
                        raise ValueError('Cannot have {} in the list of attribute \'{}\'. Expected: {} of messages of type \'{}\''.format(
                        type(t), self._priv_name[1:], imc_types[self._field_def['type']],
                        self._field_def.get('message-type', None)))

            # if it is field, check its validity, according to the IMC XML:
            if 'min' in self._field_def:
                if set_value < self._field_def['min']:
                    raise ValueError('The minimum value for attribute {} is {}. Cannot assign {}.'.format(
                        self._priv_name[1:], self._field_def['min'], set_value
                    ))

            if 'max' in self._field_def:
                if set_value > self._field_def['max']:
                    raise ValueError('The maximum value for attribute \'{}\' is {}. Cannot assign {}.'.format(
                        self._priv_name[1:], self._field_def['max'], set_value
                    ))
            
            # Check if its enumerated or bitfield
            if self._field_def.get('unit', None) == 'Enumerated':
                # Tries to get definition from the owner class. If 'enum-def' exists, it refers to a 
                # global definition; returns None, otherwise
                enum_def = self._field_def.get('enum-def', None) 
                
                # if it is global, get class from file. Else, get definition from owner class
                enum_def = getattr(imc_enums, enum_def) if enum_def else getattr(self._owner, self._priv_name[1:].upper())

                set_value = enum_def(set_value)

            if self._field_def.get('unit', None) == 'Bitfield':
                bitdef = self._field_def.get('bitfield-def', None)
                
                bitdef = getattr(imc_bitf, bitdef) if bitdef else getattr(self._owner, self._priv_name[1:].upper())

                set_value = bitdef(set_value)

            # check the size (or crop the object at serialization?)
            setattr(obj, self._priv_name, set_value)
        else:
            raise AttributeError('Cannot assign {} to {}. Expected: {}'.format(
                type(set_value), self._priv_name[1:], imc_types[self._field_def['type']]))



'''
    Extracts the IMC schema from a XML file.
'''

from . import extractutils

import xml.etree.ElementTree as ET
import pathlib

import os
import shutil
import argparse

import urllib.request
import ssl

_target_folder = 'pyimc_generated'

unknown_message = '''
class Unknown(_base_templates.base_message):
    \'\'\'A (received) message whose format is not known. It has valid sync number and valid CRC but could not be parsed.

    Contains a byte string field named 'contents' and a boolean value that indicates if its big or little endian.
    On serialization, it forces the original endianess, regardless of what the user chose to use.
    \'\'\'

    __slots__ = ['_Attributes', '_header', '_footer', '_contents', '_endianess']
    Attributes = _base_templates.immutable_attr('Message attributes. A namedtuple containing all attributes and their default values')
    contents = _base_templates.mutable_attr({'name': 'Contents', 'type': 'rawdata'}, "contents")
    endianess = _base_templates.mutable_attr({'name': 'endianess', 'type': 'rawdata'}, "endianess")

    def __init__(self, id, contents = None, endianess = None):
        \'\'\'Class constructor
        
        A (received) message whose format is not known. It has valid sync number and valid CRC but could not be parsed.

        This message class contains the following fields and their respective types:
        contents : rawdata, unit: NOT FOUND
        \'\'\'
        self._Attributes = _base_templates.MessageAttributes(fields = ('contents','endianess', ), name = "Unknown", id = id, abbrev = "Unknown", description = "A (received) message whose format is not known. It has valid sync number and valid CRC but could not be parsed.", ##ATTRIBUTES##)
        self._contents = contents
        self._endianess = endianess
    
    def _pack_fields(self, *, serial_functions : dict) -> bytes:
        raise NotImplemented

    def pack(self, *, is_field_message : bool = False, is_big_endian : bool = True, src : int = None, src_ent : int = None, 
                        dst : int = None, dst_ent : int = None) -> bytes:
        \'\'\'Serialize function that optionally overwrites the header, if parameters are provided.\'\'\'
        
        serial_functions = core.pack_functions_big if self._endianess else core.pack_functions_little
        
        s_fields = self._contents
        
        if not is_field_message:
        
            s_header = self._pack_header(serial_functions=serial_functions, size=(len(s_fields)), src=src, src_ent=src_ent, dst=dst, dst_ent=dst_ent)
            s_message = s_header + s_fields
            
            # footer:
            \'\'\'Calculates CRC-16 IBM of a bit string\'\'\'
            self._footer = core.CRC16IMB(s_message)
            s_message = s_message + serial_functions['uint16_t'](self._footer)

            return s_message
        return serial_functions['uint16_t'](self._Attributes.id) + s_fields

        '''

help_text = '''This script generates files that contain python classes as described in the IMC message schema.

Regardless of the chosen option, minimal will always be generated.

Provide a file as parameter, that contains a list of messages (abbrev's, one per line) to be white- or blacklisted.
'''

minimal = {'Abort', 'EntityState', 'QueryEntityState', 'EntityInfo', 'QueryEntityInfo', 'EntityList', 'EntityActivationState', 'QueryEntityActivationState', 
           'Heartbeat', 'Announce', 'AnnounceService'}

def hardcode_message_extractor(message : dict, templates_namespace : str, message_attributes : set) -> str:
    description = message.get('description', '')
    name = message['abbrev']
    
    namespace =  templates_namespace + '.' if templates_namespace else ''

    ws = '    '

    local_enumeration = []
    priv_attrib = ['_Attributes', '_header', '_footer']
    attributes = []
    mutable_attrib = []
    constructor_args = []
    initialization_values = []

    for attribute in message_attributes:
        
        if attribute in message:
            if attribute == 'fields':
                value = '({},)'.format(', '.join(["'{}'".format(x) for x in message[attribute].keys()]))
            else:
                value = str(message[attribute]) if isinstance(message[attribute], int) else '\"' + message[attribute].replace('"','\\"') + '\"'
        else:
            value = '[]' if attribute == 'fields' else 'None'
        attributes.append(attribute + ' = ' + value)
    
    for field in message.get('fields', []):
        priv_attrib.append('_' + field)
        initialization_values.append(2*ws + 'self.' + priv_attrib[-1] + ' = ' + field + '\n')
        constructor_args.append(field + ' = None')
        
        # All fields' contents are mutable
        field_attr = message['fields'][field]
        description = message['fields'][field].get('description', 'No description available').replace('"','\\"')
        if message['fields'][field].get('unit', None) in ['Enumerated', 'Bitfield']:
            description = description + ' ' + message['fields'][field].get('unit', None)
            if message['fields'][field].get('enum-def', None) or message['fields'][field].get('bitfield-def', None):
                description = description + ' (Global).'
            else:
                description = description + ' (Local).'

        mutable_attrib.append('{ws}{field} = {namespace}mutable_attr({definition}, \"{description}\")\n{ws}\'\'\'{description} Type: {type}\'\'\'\n'.format(
            ws = ws,
            field = field,
            namespace = namespace,
            definition = { k:v for k, v in field_attr.items() if k not in ['description', 'values']},
            description = description,
            type = message['fields'][field]['type']))

        # Determine if its a local declaration of a Enumeration/bitfield
        if (message['fields'][field].get('unit', '') == 'Enumerated' and message['fields'][field].get('enum-def', None) is None):
            
            local_enumeration.append(enum_extractor(message['fields'][field], field.upper() , False)
                                    .replace('\n', '\n' + ws))
        if (message['fields'][field].get('unit', '') == 'Bitfield' and message['fields'][field].get('bitfield-def', None) is None):
            
            local_enumeration.append(enum_extractor(message['fields'][field], field.upper(), True)
                                    .replace('\n', '\n' + ws))

    local_enumeration = ''.join(local_enumeration)
    attributes = ', '.join(attributes)
    mutable_attrib = ''.join(mutable_attrib)
    initialization_values = ''.join(initialization_values)
    constructor_args = ', '.join(constructor_args)

    fields_descriptions = []
    for x in message.get('fields', dict()):
        field_description = ws + x + ' : ' + message['fields'][x].get('type', 'NOT FOUND') + ', unit: ' + message['fields'][x].get('unit', 'NOT FOUND')
        
        if message['fields'][x].get('unit', None) in ['Enumerated', 'Bitfield']:
            if message['fields'][x].get('enum-def', None) or message['fields'][x].get('bitfield-def', None):
                field_description = field_description + ' (Global)'
            else:
                field_description = field_description + ' (Local)'

        fields_descriptions.append(field_description)
    
    description = description + '\n\n       This message class contains the following fields and their respective types:\n' + '\n\n        '.join(fields_descriptions)
    # make nested classes (Enums) immutable too?
    class_def = \
'''
class {name}({namespace}base_message):
    \'\'\'{description}\'\'\'
{local_enum}
    __slots__ = {priv_attrib}
    Attributes = {namespace}immutable_attr(\'Message attributes. A namedtuple containing all attributes and their default values\')

{mutable_attrib}
    def __init__(self, {constructor_args}):
        \'\'\'Class constructor
        
        {description}\'\'\'
        self._Attributes = {namespace}MessageAttributes({attributes})
{constructor_values}
'''.format(namespace = namespace,
name = name,
description = description,
local_enum = local_enumeration,
priv_attrib = priv_attrib,
attributes = attributes,
mutable_attrib = mutable_attrib, 
constructor_values = initialization_values,
constructor_args = constructor_args)
    
    return class_def

def enum_extractor(enum : dict, name : str, isbitfield : bool) -> str:
    '''Builds an IntEnum or IntFlag from enumerations or bitfields of the XML definition.
    Bitfields are stored as integers of powers of 2 (as expected from the definition).
    
    Obs: Though contrary to the documentation recommendation, we will be setting their
    values to support interoperability with other systems.

    Obs: If it starts with a number, it puts an 'x' at the beginning of the string.
    '''

    ws = '    '
    f_name = enum.get('name', '\"NOT FOUND\"')
    prefix = enum.get('prefix', '\"NOT FOUND\"')

    values = []
    if isbitfield:
        values.append(ws + 'EMPTY = 0\n' + ws + '\'\'\'No active flags\'\'\'\n')
    for value in enum['values']:
        if value[0].isdigit():
            values.append(ws + 'x' + value + ' = ' + str(enum['values'][value]['id']))
        else:
            values.append(ws + value + ' = ' + str(enum['values'][value]['id']))
        values.append(ws + '\'\'\'Name: ' + enum['values'][value]['name'] + '\'\'\'' + '\n')
    
    values = '\n'.join(values)
    parent_class = '_enum.IntFlag' if isbitfield else '_enum.IntEnum'
    enum_def = \
'''
class {name}({parent_class}):
    \'\'\'Full name: {f_name}
    Prefix: {prefix}\'\'\'

{values}
'''.format(name = name, parent_class = parent_class, f_name = f_name, prefix = prefix, values = values)

    return enum_def

def create_init(path):
    generated_files = os.listdir(path)

    file_name = '/__init__.py'
    with open(path + file_name, mode = 'w', encoding='utf-8') as f:
        for name in generated_files:
            if name.find('.py') > 0:
                f.write('from . import {}\n'.format(name[:name.find('.py')]))

if __name__ == '__main__':

    argparser = argparse.ArgumentParser(description=help_text)
    group = argparser.add_mutually_exclusive_group()
    group.add_argument("-w", "--whitelist", help='Messages that will be generated')
    group.add_argument("-b", "--blacklist", help='Messages that will not be generated')
    group.add_argument("-m", "--minimal", action="store_true", help='Generate minimal set only')

    args = argparser.parse_args()

    if args.whitelist is not None or args.blacklist is not None:
        file = args.whitelist if args.whitelist is not None else args.blacklist
        with open(file) as f:
            message_list = {i.strip() for i in f.readlines()}
    elif args.minimal:
        message_list = set()
    
    file = 'IMC.xml'
    try:
        tree = ET.parse(file)
        print(f'Reading {file} from current directory...')
    except FileNotFoundError:
        # Use HTTPS to get IMC.xml file from default repository
        default_repo = 'https://raw.githubusercontent.com/LSTS/imc/master/IMC.xml'
        print(f'Downloading {file} from default repository at {default_repo}.')
        response = urllib.request.urlopen(default_repo, context = ssl.create_default_context())
        IMCxml = response.read()
        print(f'Writing {file}...')
        with open(file, 'wb') as f:
            f.write(IMCxml)
        tree = ET.parse(file)
    
    print('Extracting messages...')

    root = tree.getroot()

    # Split the XML into metadata and messages
    raw_metadata = [x for x in root if x.tag != 'message']
    raw_messages = [x for x in root if x.tag == 'message']

    metadata_encyclopedia = {x.tag : extractutils.recursive_parser(x) for x in raw_metadata}
    metadata_encyclopedia = extractutils.tree_shortener(metadata_encyclopedia,'')

    message_encyclopedia = {int(x.attrib['id']) : extractutils.recursive_parser(x) for x in raw_messages}

    if args.blacklist is not None:
        message_encyclopedia = {k: v for k,v in message_encyclopedia.items() if v['abbrev'] not in [i for i in message_list if i not in minimal]}
    elif args.whitelist is not None or args.minimal:
        message_encyclopedia = {k: v for k,v in message_encyclopedia.items() if v['abbrev'] in message_list.union(minimal)}

    # Automatically built as a list, this key in particular is needed as a dict of lists of ids, but
    # we have to manually reshape/calculate it
    metadata_encyclopedia['categories'] = {cat: [] for cat in metadata_encyclopedia['categories']}
    for (id, msg) in message_encyclopedia.items():
        category = msg.get('category', None)
        metadata_encyclopedia['categories'][category].append((msg.get('id')))
    
    # Hardcode IMC types to python types
    imc_types = dict()
    for t in metadata_encyclopedia['types']['types']:
        # "Deduce" types by their description.
        imc_types[t] =  float if 'float' in metadata_encyclopedia['types']['types'][t]['description'] else \
                        int if 'integer' in metadata_encyclopedia['types']['types'][t]['description'] else \
                        None
    imc_types['rawdata'] = bytes
    imc_types['plaintext'] = str
    imc_types['message'] = '<class \'IMC_message\'>'
    imc_types['message-list'] = list

    # XML variables inspection: Get all possible attributes.
    message_attributes = set()
    for message in message_encyclopedia:
        for attrib in message_encyclopedia[message]:
            message_attributes.add(attrib.replace('-',''))

    fields_attributes = set()
    for message in message_encyclopedia:
        for attrib in message_encyclopedia[message].get('fields', []):
            for child_attrib in message_encyclopedia[message]['fields'][attrib]:
                fields_attributes.add(child_attrib)

    # future flag
    force = True

    if os.path.isdir(_target_folder):
        if os.listdir(_target_folder):
            if force:
                shutil.rmtree(_target_folder)
                os.mkdir(_target_folder)
            else:
                exit('Folder \'{}\' is not empty!'.format(_target_folder))
    else:
        os.mkdir(_target_folder)
    
    os.mkdir(_target_folder + '/categories')

    # Write files
    file_name = 'enumerations.py'
    with open(_target_folder + '/' + file_name, mode = 'w', encoding='utf-8') as f:
        f.write('\'\'\'\nIMC global enumerations definitions.\n\'\'\'\n\n')
        f.write('import enum as _enum\n\n#Enumerations:\n')
        for k, v in metadata_encyclopedia['enumerations'].items():
            f.write(enum_extractor(v, k, False))
    
    file_name = 'bitfields.py'
    with open(_target_folder + '/' + file_name, mode = 'w', encoding='utf-8') as f:
        f.write('\'\'\'\nIMC global bitfields definitions.\n\'\'\'\n\n')
        f.write('import enum as _enum\n\n#Enumerations:\n')
        for k, v in metadata_encyclopedia['bitfields'].items():
            f.write(enum_extractor(v, k, True))
    
    lib_location = pathlib.Path(__file__).parent.resolve()
    file_name = '_base_templates.py'
    with open(str(lib_location) + '/' + file_name, mode = 'r', encoding='utf-8') as f_in:
        base_templates = f_in.read()
        base_templates = base_templates.replace('%SYNCH_NUMBER%', hex(metadata_encyclopedia['header']['fields']['sync']['value']))
        base_templates = base_templates.replace('%IMC_TYPES%', str(imc_types).replace('<class \'', '').replace("'>",'').replace('"',''))
        base_templates = base_templates.replace('%MESSAGE_ATTRIBUTES%', str([s.replace('-','') for s in message_attributes]))

        with open(_target_folder + '/' + file_name, mode = 'w', encoding='utf-8') as f_out:           
            f_out.write(base_templates)

    file_name = 'messages.py'
    with open(_target_folder + '/' + file_name, mode = 'w', encoding='utf-8') as f:
        f.write('\'\'\'\nIMC messages.\n\'\'\'\n\n')
        # write import statements
        f.write('from . import _base_templates\nimport enum as _enum\nimport pyimclsts.core as core\nfrom . import categories\n')
        f.write('\n_message_ids = {}\n'.format(str(dict((k, v['abbrev']) for k, v in message_encyclopedia.items()))))
        f.write(unknown_message.replace('##ATTRIBUTES##', ', '.join([i + '= None' for i in message_attributes if i not in {'fields', 'name', 'id', 'abbrev', 'description'}])))

        messages_w_cat = []
        for cat, l in metadata_encyclopedia['categories'].items():
            l_filtered = [x for x in l if x in message_encyclopedia.keys()]
            if l_filtered:
                with open(_target_folder + '/categories/' + cat.replace(' ', '') + '.py', mode = 'w', encoding='utf-8') as f_cat:
                    f_cat.write(f'\'\'\'\nIMC {cat} messages.\n\'\'\'\n\n')
                    # write import statements
                    f_cat.write('from .. import _base_templates\nimport enum as _enum\nimport pyimclsts.core as core\n')
                    
                    for id in l_filtered:
                        f_cat.write(hardcode_message_extractor(message_encyclopedia[id], '_base_templates', message_attributes))    
                messages_w_cat = messages_w_cat + l_filtered
        
        for k, v in message_encyclopedia.items():
            if k in messages_w_cat:
                f.write('\n\n' + v['abbrev'] + ' = categories.' + v['category'].replace(' ', '') + '.' + v['abbrev'])
            else:
                f.write(hardcode_message_extractor(v, '_base_templates', message_attributes))
    
    create_init(_target_folder + '/categories')
    create_init(_target_folder)

    print('Finished extracting messages.')



'''
    Contains functions to help develop/build the IMC binding, but are not used
    after the binding is complete.
'''
import xml.etree.ElementTree as ET

def recursive_parser(element):
    #print(element.attrib.get('id', None))
    node = {}
    if element.tag == 'description':
        # Check if it is empty... It happens =/
        if element.text:  # not empty
            # Clean the string (remove double whitespaces and line breaks)
            node[element.tag] = " ".join(element.text.split())
        else:
            node[element.tag] = ""
        return node

    # Operation over every node
    for a in element.attrib:
        if element.attrib[a].isdigit() or element.attrib[a].startswith('0x'):
            node[a] = int(element.attrib[a], 0)
        else:
            try: # Jesus Christ, forgive me for I have sinned
                node[a] = float(element.attrib[a])
            except ValueError:
                node[a] = element.attrib[a]
            
    # if the tag of child node belongs to a dummy_tag, merge them 
    # into the parent's name key (xml tag), preserving each entry
    children = []
    for child in element:
        if child.tag == 'description':
            # Check if it is empty... It happens =/
            if child.text: # not empty
                # Clean the string (remove double whitespaces and line breaks)
                node[child.tag] = " ".join(child.text.split())
            else:
                node[child.tag] = ""
        else:
            children.append(recursive_parser(child))

    # make a dictionary indexed by the abbrev or the name of the attributes
    dict_children = {}
    for child in children:
        key = child.get('abbrev', None)
        if not key:
            name = child.get('name', None)
            del child['name']
        else:
            name = child.get('abbrev', None)
            del child['abbrev']
        
        dict_children[name] = child
        
    # if the children values are empty, use a list instead
    if not all(dict_children.values()):
        dict_children = list(dict_children)

    # check if its an empty list to avoid having an empty key
    if dict_children:
        # fetch the common tag among the children. (They are always(?) the same, except for 'description')
        new_tag = list(filter(lambda x : x.tag != 'description', list(element)))[0].tag
        node[new_tag + 's'] = dict_children     # shameless exploit of English morphology
        #node[element.tag] = dict_children

    return node

def recursive_print(node : dict, ws : str = '' ):
    ws_ = "  " + ws
    for key in node.keys():
        if isinstance(node[key], dict):
            print(ws_ + str(key) + ' : ')
            recursive_print(node[key], ws_)
        else:
            print(ws_ + key + ' : ' + str(node[key]))

def tree_shortener(tree : dict, name : any) -> dict:
    '''Removes tree levels that consists of a single node.'''
    new_tree = dict()

    if isinstance(tree, dict):
        if len(tree) == 1:
            key = list(tree.keys())[0]
            
            if key == 'description': # leave this one untouched.
                return tree

            new_tree = tree[key]
            return new_tree
        
        for (key, value) in tree.items():
            new_tree[key] = tree_shortener(value, key)
    else:
        return tree

    return new_tree

if __name__ == '__main__':
    file = 'IMC.xml'

    tree = ET.parse(file)
    root = tree.getroot()

    # Split the XML into metadata and messages
    raw_metadata = [x for x in root if x.tag != 'message']
    raw_messages = [x for x in root if x.tag == 'message']

    # test variables
    x = raw_metadata[1]
    y = x[1]

    metadata_encyclopedia = {x.tag : recursive_parser(x) for x in raw_metadata}
    message_catalog = {int(x.attrib['id']) : recursive_parser(x) for x in raw_messages}

    recursive_print(message_catalog[1])
    recursive_print(metadata_encyclopedia['types'])
    d = filter(lambda x: x['id'] == 552, message_catalog)